#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017 jfrfonseca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Operacoes com arquivos ZIP"""


"""
# IMPORTS
"""


# Biblioteca padrão
import zipfile
import tempfile


"""
# UTILS
"""


def le_arquivo_zip_de_string(string_zipada):
    arquivo_temporario = tempfile.TemporaryFile()
    arquivo_temporario.write(string_zipada)
    string_zipada_zip = zipfile.ZipFile(arquivo_temporario, "r")
    return string_zipada_zip.read(string_zipada_zip.filelist[0].filename)
